import golly as g 
import os
import inspect

LONG_ENOUGH = 8192
CIRCUIT_RECOVERY = 14
CHECK_SPECIAL_CASES = False  #True

filename = inspect.getframeinfo(inspect.currentframe()).filename
scriptpath = os.path.dirname(os.path.abspath(filename))
fname = "simeks-library.txt"
outfname1 = "optimized.txt"
outfname2 = "adjustable.txt"

gliders=["3o$o$bo!","b2o$2o$2bo!","b2o$bobo$bo!","2bo$b2o$bobo!"]
gliderlist=[g.parse(gl) for gl in gliders]
elbow=g.parse("2o$2o!")

g.setrule("Life")

def makerecipe(recipe):
  clist = gliderlist[0]
  paritytype = recipe[0]%2 # the first number in the list is a parity indicator:  0=any, 1=odd, 2=even
  timings=[paritytype]

  for i in recipe[1:]:
    newglider=timings[-1]+i
    timings+=[newglider]
    clist=g.join(clist,g.transform(gliderlist[newglider%4],newglider/4,newglider/4))
    # TODO:  not actually doing anything with timings, for the moment
  return clist

def testrecipe(recipe):
  pat=makerecipe(recipe)
  g.addlayer()
  g.putcells(g.join(elbow,g.transform(pat,5,2)))
  g.run(LONG_ENOUGH)
#  r=g.getrect()
#  if r[0]<-128: r[0]=-128
#  if r[1]<-128: r[1]=-128 # moves the rectangle to exclude any NE/SW/NW output gliders
  g.setcell(-128,-128,1) # this ensures that the position of the output ash and glider will matter
  result=g.hash([-128,-128,256,256])
  g.dellayer()
  return result

with open(os.path.join(scriptpath,fname),'r') as f: data=f.readlines()

with open(os.path.join(scriptpath,outfname1),'w') as outf1:
  outf1.write("Shortest versions of recipes:\n")
with open(os.path.join(scriptpath,outfname2),'w') as outf2:
  outf2.write("Recipes showing adjustment options:\n")


count=0
for item in data:
  count+=1
  newrecipe="[0,"
  if item[:1]=="#": continue
  pieces=item.replace("/* ","").replace(", 0},","").split(" */  {")
  name=pieces[0]
  offsetlist=[int(i) for i in pieces[1].split(",")]
  
  offsetlist+=[512,16]  # add trailing gliders, to find out what the closest
                        # safe following distance is for the next recipe
  
  standardoutput=testrecipe(offsetlist)
  for index in range(1,len(offsetlist)-1): # skip the parity value at the beginning of the recipe
                                           # and also the "16" trailing cleanup glider added at the end
    min=offsetlist[index]
    origmin=min
    newmin=min+2
    newlist = offsetlist[:index]+[newmin]+offsetlist[index+1:]
    if standardoutput==testrecipe(newlist):
      # if it's slower by two ticks, the recipe still works!  That means this glider can be adjusted...
      newmin=origmin
      low_bdy=CIRCUIT_RECOVERY
      high_bdy=min
      halfway=int((high_bdy-low_bdy)/2)
      # binary search --
      #   may miss a non-contiguous lower stretch of valid timings, but probably wont
      while halfway>0:
        steps=int((halfway+1)/2)
        newmin = high_bdy - steps*2 # they're 2-tick steps -- it's a little confusing
        newlist = offsetlist[:index]+[newmin]+offsetlist[index+1:]
        g.show(str(count) + ":" + str(newlist))
        if standardoutput==testrecipe(newlist):
          min=newmin
          high_bdy=newmin
        else:
          low_bdy+=2
          if low_bdy<newmin: low_bdy=newmin
        halfway=int((high_bdy-low_bdy)/2)
      note="+2N"
      
      # Now see if this is a P2 or P1 adjustable glider
      newmin=min+1
      newlist = offsetlist[:index]+[newmin]+offsetlist[index+1:]
      if standardoutput==testrecipe(newlist):
        note="+N"

      # Now check the only remaining boundary case, if there is one
      newmin-=2
      if newmin>=CIRCUIT_RECOVERY:
        newlist = offsetlist[:index]+[newmin]+offsetlist[index+1:]
        if standardoutput==testrecipe(newlist): min=newmin

    else:
      note=""  # this glider is not adjustable

    note=str(min)+note
    # optional:  record any special cases where the range of allowable timings is not contiguous
    if CHECK_SPECIAL_CASES:
      newmin=min
      while newmin>CIRCUIT_RECOVERY:
        newmin-=1
        newlist = offsetlist[:index]+[newmin]+offsetlist[index+1:]
        g.show(str(count) + ":" + str(newlist) + " -- double-checking to recovery-time boundary...")
        if standardoutput==testrecipe(newlist) and newmin<min:
          note+="|"+str(newmin)
          min=newmin    

    newrecipe+=note+","
    # If there's no output glider, then we might as well use the new minimum recipe
    # -- but for now there's no way to tell this from the library file, so we'll
    # play it safe and leave the original offsetlist alone.
    # TODO:  Find and delete the output glider before hashing,
    #        and figure out a safe place to drop the single cell
    #        (or leave out the single cell, since it's unlikely to matter.)
    #        Otherwise the optimization fails because gliders in the
    #        recipe before the output glider appears, can't be moved.
    offsetlist[index]=min

  with open(os.path.join(scriptpath,outfname1),'a') as outf1: outf1.write(name+":"+str(offsetlist[:-1])+"\n")
  with open(os.path.join(scriptpath,outfname2),'a') as outf2: outf2.write(name+":"+newrecipe[:-1]+"]\n")
  pat=makerecipe(offsetlist[:-1])
  g.addlayer()
  g.putcells(g.join(elbow,g.transform(pat,5,2)))

  
g.show("Optimization complete.")